@@ -2,7 +2,7 @@
  *	Copyright 1996, University Corporation for Atmospheric Research
  *      See netcdf/COPYRIGHT file for copying and redistribution conditions.
  */
-/* $Id: nc.c,v 2.118 2001/01/11 20:39:09 russ Exp $ */
+/* $Id: nc.c,v 2.135 2004/09/30 18:47:12 russ Exp $ */
 
 #include "nc.h"
 #include "rnd.h"
@@ -18,6 +18,15 @@
 /* list of open netcdf's */
 static NC *NClist = NULL;
 
+/* This is the default create format for nc_create and nc__create. */
+static int default_create_format = NC_FORMAT_CLASSIC;
+
+/* These have to do with version numbers. */
+#define MAGIC_NUM_LEN 4
+#define VER_CLASSIC 1
+#define VER_64BIT_OFFSET 2
+#define VER_HDF5 3
+
 static void
 add_to_NCList(NC *ncp)
 {
@@ -107,7 +116,7 @@ new_NC(const size_t *chunkp)
 	(void) memset(ncp, 0, sizeof(NC));
 
 	ncp->xsz = MIN_NC_XSZ;
-	assert(ncp->xsz == ncx_len_NC(ncp));
+	assert(ncp->xsz == ncx_len_NC(ncp,0));
 	
 	ncp->chunk = chunkp != NULL ? *chunkp : NC_SIZEHINT_DEFAULT;
 
@@ -200,12 +209,13 @@ ncx_howmany(nc_type type, size_t xbufsize)
  * Compute each variable's 'begin' offset,
  * update 'begin_rec' as well.
  */
-static void
+static int
 NC_begins(NC *ncp,
 	size_t h_minfree, size_t v_align,
 	size_t v_minfree, size_t r_align)
 {
 	size_t ii;
+	int sizeof_off_t;
 	off_t index = 0;
 	NC_var **vpp;
 	NC_var *last = NULL;
@@ -215,10 +225,16 @@ NC_begins(NC *ncp,
 	if(r_align == NC_ALIGN_CHUNK)
 		r_align = ncp->chunk;
 
-	ncp->xsz = ncx_len_NC(ncp);
+	if (fIsSet(ncp->flags, NC_64BIT_OFFSET)) {
+	  sizeof_off_t = 8;
+	} else {
+	  sizeof_off_t = 4;
+	}
+	
+	ncp->xsz = ncx_len_NC(ncp,sizeof_off_t);
 
 	if(ncp->vars.nelems == 0) 
-		return;
+		return NC_NOERR;
 
 	/* only (re)calculate begin_var if there is not sufficient space in header
 	   or start of non-record variables is not aligned as requested by valign */
@@ -246,6 +262,10 @@ NC_begins(NC *ncp,
 #if 0
 fprintf(stderr, "    VAR %d %s: %ld\n", ii, (*vpp)->name->cp, (long)index);
 #endif
+                if( sizeof_off_t == 4 && (index > X_OFF_MAX || index < 0) ) 
+		{
+		    return NC_EVARSIZE;
+                }
 		(*vpp)->begin = index;
 		index += (*vpp)->len;
 	}
@@ -279,8 +299,19 @@ fprintf(stderr, "    VAR %d %s: %ld\n", ii, (*vpp)->name->cp, (long)index);
 #if 0
 fprintf(stderr, "    REC %d %s: %ld\n", ii, (*vpp)->name->cp, (long)index);
 #endif
+                if( sizeof_off_t == 4 && (index > X_OFF_MAX || index < 0) ) 
+		{
+		    return NC_EVARSIZE;
+                }
 		(*vpp)->begin = index;
 		index += (*vpp)->len;
+		/* check if record size must fit in 32-bits */
+#if SIZEOF_OFF_T == SIZEOF_SIZE_T && SIZEOF_SIZE_T == 4
+		if( ncp->recsize > X_UINT_MAX - (*vpp)->len )
+		{
+		    return NC_EVARSIZE;
+		}
+#endif
 		ncp->recsize += (*vpp)->len;
 		last = (*vpp);
 	}
@@ -293,7 +324,7 @@ fprintf(stderr, "    REC %d %s: %ld\n", ii, (*vpp)->name->cp, (long)index);
 
 	if(NC_IsNew(ncp))
 		NC_set_numrecs(ncp, 0);
-
+	return NC_NOERR;
 }
 
 
@@ -305,7 +336,7 @@ int
 read_numrecs(NC *ncp)
 {
 	int status = NC_NOERR;
-	const void *xp;
+	const void *xp = NULL;
 	size_t nrecs = NC_get_numrecs(ncp);
 
 	assert(!NC_indef(ncp));
@@ -340,7 +371,7 @@ int
 write_numrecs(NC *ncp)
 {
 	int status = NC_NOERR;
-	void *xp;
+	void *xp = NULL;
 
 	assert(!NC_readonly(ncp));
 	assert(!NC_indef(ncp));
@@ -627,6 +658,88 @@ move_vars_r(NC *gnu, NC *old)
 }
 
 
+/*
+ * Given a valid ncp, return NC_EVARSIZE if any variable has a bad len 
+ * (product of non-rec dim sizes too large), else return NC_NOERR.
+ */
+static int
+NC_check_vlens(NC *ncp)
+{
+    NC_var **vpp;
+    /* maximum permitted variable size (or size of one record's worth
+       of a record variable) in bytes.  This is different for format 1
+       and format 2. */
+    size_t vlen_max;
+    size_t ii;
+    size_t large_vars_count;
+    size_t rec_vars_count;
+    int last;
+
+    if(ncp->vars.nelems == 0) 
+	return NC_NOERR;
+
+    if ((ncp->flags & NC_64BIT_OFFSET) && sizeof(off_t) > 4) {
+	/* CDF2 format and LFS */
+	vlen_max = X_UINT_MAX - 3; /* "- 3" handles rounded-up size */
+    } else {
+	/* CDF1 format */
+	vlen_max = X_INT_MAX - 3;
+    }
+    /* Loop through vars, first pass is for non-record variables.   */
+    large_vars_count = 0;
+    rec_vars_count = 0;
+    vpp = ncp->vars.value;
+    for (ii = 0; ii < ncp->vars.nelems; ii++, vpp++) {
+	if( !IS_RECVAR(*vpp) ) {
+	    last = 0;
+	    if( NC_check_vlen(*vpp, vlen_max) == 0 ) {
+		large_vars_count++;
+		last = 1;
+	    }
+	} else {
+	  rec_vars_count++;
+	}
+    }
+    /* OK if last non-record variable size too large, since not used to 
+       compute an offset */
+    if( large_vars_count > 1) { /* only one "too-large" variable allowed */
+      return NC_EVARSIZE;
+    }
+    /* and it has to be the last one */ 
+    if( large_vars_count == 1 && last == 0) { 
+      return NC_EVARSIZE;
+    }
+    if( rec_vars_count > 0 ) {
+	/* and if it's the last one, there can't be any record variables */
+	if( large_vars_count == 1 && last == 1) {
+	    return NC_EVARSIZE;
+	}
+	/* Loop through vars, second pass is for record variables.   */
+	large_vars_count = 0;
+	vpp = ncp->vars.value;
+	for (ii = 0; ii < ncp->vars.nelems; ii++, vpp++) {
+	    if( IS_RECVAR(*vpp) ) {
+		last = 0;
+		if( NC_check_vlen(*vpp, vlen_max) == 0 ) {
+		    large_vars_count++;
+		    last = 1;
+		}
+	    }
+	}
+	/* OK if last record variable size too large, since not used to 
+	   compute an offset */
+	if( large_vars_count > 1) { /* only one "too-large" variable allowed */
+	    return NC_EVARSIZE;
+	}
+	/* and it has to be the last one */ 
+	if( large_vars_count == 1 && last == 0) { 
+	    return NC_EVARSIZE;
+	}
+    }
+    return NC_NOERR;
+}
+
+
 /*
  *  End define mode.
  *  Common code for ncendef, ncclose(endef)
@@ -642,7 +755,12 @@ NC_endef(NC *ncp,
 	assert(!NC_readonly(ncp));
 	assert(NC_indef(ncp));
 
-	NC_begins(ncp, h_minfree, v_align, v_minfree, r_align);
+	status = NC_check_vlens(ncp);
+	if(status != NC_NOERR)
+	    return status;
+	status = NC_begins(ncp, h_minfree, v_align, v_minfree, r_align);
+	if(status != NC_NOERR)
+	    return status;
 
 	if(ncp->old != NULL)
 	{
@@ -748,6 +866,7 @@ nc__create_mp(const char * path, int ioflags, size_t initialsz, int basepe,
 	NC *ncp;
 	int status;
 	void *xp = NULL;
+	int sizeof_off_t = 0;
 
 #if ALWAYS_NC_SHARE /* DEBUG */
 	fSet(ioflags, NC_SHARE);
@@ -768,7 +887,21 @@ nc__create_mp(const char * path, int ioflags, size_t initialsz, int basepe,
 	if(basepe != 0)
 		return NC_EINVAL;
 #endif
-	assert(ncp->xsz == ncx_len_NC(ncp));
+
+	assert(ncp->flags == 0);
+
+	/* Apply default create format. */
+	if (default_create_format == NC_FORMAT_64BIT)
+	  ioflags |= NC_64BIT_OFFSET;
+
+	if (fIsSet(ioflags, NC_64BIT_OFFSET)) {
+	  fSet(ncp->flags, NC_64BIT_OFFSET);
+	  sizeof_off_t = 8;
+	} else {
+	  sizeof_off_t = 4;
+	}
+
+	assert(ncp->xsz == ncx_len_NC(ncp,sizeof_off_t));
 	
 	status = ncio_create(path, ioflags,
 		initialsz,
@@ -782,7 +915,6 @@ nc__create_mp(const char * path, int ioflags, size_t initialsz, int basepe,
 		goto unwind_alloc;
 	}
 
-	assert(ncp->flags == 0);
 	fSet(ncp->flags, NC_CREAT);
 
 	if(fIsSet(ncp->nciop->ioflags, NC_SHARE))
@@ -797,7 +929,7 @@ nc__create_mp(const char * path, int ioflags, size_t initialsz, int basepe,
 		fSet(ncp->flags, NC_NSYNC);
 	}
 
-	status = ncx_put_NC(ncp, &xp, 0, ncp->xsz);
+	status = ncx_put_NC(ncp, &xp, sizeof_off_t, ncp->xsz);
 	if(status != NC_NOERR)
 		goto unwind_ioc;
 
@@ -817,13 +949,31 @@ unwind_alloc:
 	return status;
 }
 
+/* This function sets a default create flag that will be logically
+   or'd to whatever flags are passed into nc_create for all future
+   calls to nc_create.
+   Valid default create flags are NC_64BIT_OFFSET, NC_CLOBBER,
+   NC_LOCK, NC_SHARE. */
+int
+nc_set_default_format(int format, int *old_formatp)
+{
+    /* Return existing format if desired. */
+    if (old_formatp)
+      *old_formatp = default_create_format;
+
+    /* Make sure only valid format is set. */
+    if (format != NC_FORMAT_CLASSIC && format != NC_FORMAT_64BIT)
+      return NC_EINVAL;
+    default_create_format = format;
+    return NC_NOERR;
+}
+
 int
 nc_create(const char * path, int ioflags, int *ncid_ptr)
 {
 	return nc__create(path, ioflags, 0, NULL, ncid_ptr);
 }
 
-
 int
 nc__open(const char * path, int ioflags,
 	size_t *chunksizehintp, int *ncid_ptr)
@@ -858,6 +1008,7 @@ nc__open_mp(const char * path, int ioflags, int basepe,
 	if(basepe != 0)
 		return NC_EINVAL;
 #endif
+
 	status = ncio_open(path, ioflags,
 		0, 0, &ncp->chunk,
 		&ncp->nciop, 0);
@@ -1098,7 +1249,7 @@ nc_redef(int ncid)
 	if(fIsSet(ncp->nciop->ioflags, NC_SHARE))
 	{
 		/* read in from disk */
-		status = NC_sync(ncp);
+		status = read_NC(ncp);
 		if(status != NC_NOERR)
 			return status;
 	}
